"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json

from json.decoder import JSONDecodeError
from logging import Logger
from jsonschema import Draft7Validator
from .Exceptions.IOException import IOException
from .Exceptions.JSONException import JSONException
from .Exceptions.JSONSchemaException import JSONSchemaException
from .Logging import Logging


class Config:
	"""Configuration class."""

	@staticmethod
	def read(config_path: str, schema_path: str) -> dict:
		"""
		Read and validate configuration file against JSON schema, then return configuration.

		Parameters:
			config_path (str): Path to configuration file
			schema_path (str): Path to configuration jsonschema file
		Returns:
			config (dict): Configuration dict
		Exceptions:
			IOException: raised if configuration or schema file does not exist
			JSONException: raised if configuration or schema file have incorrect JSON syntax
			JSONSchemaException: raised if configuration file does not validate against configuration schema
		"""
		logger: Logger = Logging.get_logger()
		try:
			path = config_path
			with open(path, mode='r', encoding='utf-8') as file:
				config = json.load(file)

			path = schema_path
			with open(schema_path, mode='r', encoding='utf-8') as file:
				schema = json.load(file)

			validator = Draft7Validator(schema)
			validation_errors = sorted(validator.iter_errors(config), key=str)

			if validation_errors:
				for error in validation_errors:
					logger.error(error)
				raise JSONSchemaException(f'Failed to validate configuration file {config_path} against schema {schema_path}.')
			return config
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			logger.error(error_str)
			raise IOException(error_str) from io_error
		except JSONDecodeError as json_error:
			error_str = Logging.json_file_error(json_error, path)
			logger.error(error_str)
			raise JSONException(error_str) from json_error
