"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from logging import Logger
from typing import List
from influxdb_client import InfluxDBClient, Point
from influxdb_client.client.write_api import WriteApi
from influxdb_client.domain.write_precision import WritePrecision
from .Entities.Measurement import Measurement
from .InfluxClientInterface import InfluxClientInterface
from .Logging import Logging


class InfluxClient(InfluxClientInterface):
	"""
	Influxdb (1.8/2.0+) client class.

	Attributes
	----------
	logger: Logger
		logger object
	host: str
		influxdb hostname
	port: int
		influxdb port
	token: str
		influxdb api access token
	bucket: str
		influxdb database name
	org: str
		influxdb organization name
	client: InfluxDBClient
		influxdb client

	Methods
	-------
	start() -> None:
		Initializes Influxdb client and connects to database.
	write(measurements: List[Measurement]) -> None:
		Writes measurements to database and logs information.
	"""

	def __init__(self, config: dict) -> None:
		"""
		Construct attributes for Influxdb client object.

		Parameters
		----------
		config: dict
			influxdb client configuration dictionary

		Returns
		-------
		None
		"""
		self.logger: Logger = Logging.get_logger()
		self.host: str = config['host']
		self.port: int = config['port']
		self.token: str = config['token']
		self.bucket: str = config['bucket']
		self.org: str = config['org']
		self.client: InfluxDBClient = None
		self.writer: WriteApi = None

	def start(self) -> None:
		"""
		Initialize Influxdb client and connects to database.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		self.client = InfluxDBClient(
			url=f'http://{self.host}:{self.port}',
			token=self.token,
			org=self.org
		)
		self.writer = self.client.write_api()

	def write(self, measurements: List[Measurement]) -> None:
		"""
		Write measurements to database and logs information.

		Parameters
		----------
		measurements: List[Measurement]
			list of measurement objects to be written to database

		Returns
		-------
		None
		"""
		for measurement in measurements:
			record: Point = Point.from_dict(measurement.to_dict())
			self.writer.write(bucket=self.bucket, write_precision=WritePrecision.NS, record=record)
			self.logger.info(Logging.log_measurement(measurement))
