"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from logging import Logger
from typing import List
from influxdb import InfluxDBClient
from .Entities.Measurement import Measurement
from .InfluxClientInterface import InfluxClientInterface
from .Logging import Logging


class InfluxClientLegacy(InfluxClientInterface):
	"""
	Legacy Influxdb (1.x) client class.

	Attributes
	----------
	logger: Logger
		logger object
	host: str
		influxdb hostname
	port: int
		influxdb port
	user: str
		influxdb client access username
	password: str
		inflxudb client access password
	bucket: str
		influxdb database name
	org: str
		influxdb organization name
	client: InfluxDBClient
		influxdb client

	Methods
	-------
	start() -> None:
		Initializes Influxdb client and connects to database.
	write(measurements: List[Measurement]) -> None:
		Writes measurements to database and logs information.
	"""

	# pylint: disable=too-many-instance-attributes

	def __init__(self, config: dict) -> None:
		"""
		Construct attributes for legacy Influxdb client object.

		Parameters
		----------
		config: dict
			influxdb client configuration dictionary

		Returns
		-------
		None
		"""
		self.logger: Logger = Logging.get_logger()
		self.host: str = config['host']
		self.port: int = config['port']
		self.user: str = config['user']
		self.password: str = config['password']
		self.bucket: str = config['bucket']
		self.org: str = config['org']
		self.client: InfluxDBClient = None

	def start(self) -> None:
		"""
		Initialize Influxdb client and connects to database.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		self.client = InfluxDBClient(
			host=self.host,
			port=self.port,
			username=self.user,
			password=self.password,
			database=self.bucket
		)

	def write(self, measurements: List[Measurement]) -> None:
		"""
		Write measurements to database and logs information.

		Parameters
		----------
		measurements: List[Measurement]
			list of measurement objects to be written to database

		Returns
		-------
		None
		"""
		points = [measurement.to_dict() for measurement in measurements]
		self.client.write_points(points, time_precision='n')
