"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from json.decoder import JSONDecodeError
from logging import handlers, Logger, _nameToLevel as LogNameToLevel
from systemd.journal import JournalHandler
from .Entities.Measurement import Measurement


class Logging:
	"""Logging configuration, logger fetch and error formatting class."""

	__log_path = '/var/log/iqrf-gateway-influxdb-bridge/iqrf-gateway-influxdb-bridge.log'
	__file_format = '[%(asctime)s]: [%(filename)s:%(lineno)s]: [%(levelname)s] %(message)s'
	__journal_format = '[%(levelname)s] %(message)s'
	__log_level = logging.INFO

	@classmethod
	def set_level(cls, level: str) -> None:
		"""
		Set logging level.

		Parameters
		----------
		level: str
			Logging level.

		Returns
		-------
		None
		"""
		cls.__log_level = cls.level_name_to_value(level)

	@classmethod
	def get_logger(cls) -> Logger:
		"""
		Initialize and return a new logger object.

		Parameters
		----------
		None

		Returns
		-------
		Logger
			Logger object.
		"""
		logger = logging.getLogger(__name__)
		if not logger.hasHandlers():
			logger.propagate = False
			file_handler = handlers.RotatingFileHandler(cls.__log_path, maxBytes=4000000, backupCount=5, mode='a')
			file_handler.setFormatter(logging.Formatter(cls.__file_format))
			journal_handler = JournalHandler()
			journal_handler.setFormatter(logging.Formatter(cls.__journal_format))
			logger.addHandler(file_handler)
			logger.addHandler(journal_handler)
			logger.setLevel(cls.__log_level)
		return logger

	@staticmethod
	def level_name_to_value(level: str) -> int:
		"""
		Convert log level name to value.

		Parameters
		----------
		level: str
			log level name

		Returns
		-------
		int
			log level numeric value
		"""
		return LogNameToLevel[level]

	@staticmethod
	def log_measurement(measurement: Measurement) -> str:
		"""
		Create and return database insert log string.

		Parameters
		----------
		measurement: Measurement
			measurement object

		Returns
		-------
		str
			Database insert log string.
		"""
		name = measurement.name
		gwid = measurement.device.gwid
		addr = measurement.device.address
		idx = measurement.device.index
		value = measurement.value
		return f'Added to db measurement: {name}, gwid: {gwid}, nadr: {addr}, idx: {idx}, value: {value}.'

	@staticmethod
	def io_error(error: IOError) -> str:
		"""
		Create and return IO error log string.

		Parameters
		----------
		error: IOError
			io exception object

		Returns
		-------
		str
			IO error log string.
		"""
		return f'{error.strerror}: {error.filename}'

	@staticmethod
	def json_response_parse_error(error: JSONDecodeError) -> str:
		"""
		Create and return JSON message parsing error log string.

		Parameters
		----------
		error: JSONDecodeError
			json decoding error

		Returns
		-------
		str
			JSON message parsing error log string.
		"""
		return f'Failed to parse message. {str(error)}. Received message: {error.doc}'

	@staticmethod
	def json_file_error(error: JSONDecodeError, file: str) -> str:
		"""
		Create and return JSON file validation error log string.

		Parameters
		----------
		error: JSONDecodeError
			json decoding error
		file: str
			path to validated file

		Returns
		-------
		str
			JSON file validation error log string.
		"""
		return f'{error.msg} at line {error.lineno} column {error.colno} (char {error.pos}) in file {file}.'

	@staticmethod
	def json_request_error(error: JSONDecodeError, topic: str) -> str:
		"""
		Create and return JSON request validation error log string.

		Parameters
		----------
		error: JSONDecodeError
			json decoding error
		topic: str
			mqtt response topic

		Returns
		-------
		str
			JSON request validation error log string.
		"""
		return f'{error.msg} at line {error.lineno} column {error.colno} (char {error.pos}) in message received at topic {topic}.'
