"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Tuple, Union

from .Enums.ChargerData import ChargerData
from .Enums.Messages import Messages
from .Entities.DeviceInfo import DeviceInfo
from .Entities.Measurement import Measurement
from .Exceptions.MessageException import MessageException
from .Exceptions.SensorExceptions import SensorException


class MessageParser:
	"""Collection of functions for parsing of incoming messages."""

	@staticmethod
	def parse_message(message: dict, gwid: str) -> List[Measurement]:
		"""
		Check message execution status code and parse message.

		If response status code is not errorrous, returns list of measurement objects.

		Parameters
		----------
		message: dict
			response message
		gwid: str
			gateway id

		Returns
		-------
		List[Measurement]
			list of measurement objects

		Exceptions
		----------
		MessageException
			raised if response status code indicates error
		"""
		message_type = message['mType']
		message_id = message['data']['msgId']
		status = message['data']['status']
		if status != 0:
			if 'statusStr' in message['data']:
				raise MessageException(f'Request {message_type} ({message_id}) failed with error {status} ({message["data"]["statusStr"]}).')
			raise MessageException(f'Request {message_type} ({message_id}) failed with error {status}.')
		if message_type == Messages.MCU_CHARGER:
			return MessageParser.parse_charger(message['data']['rsp'], gwid)
		if message_type in [Messages.FRC_SEND, Messages.FRC_SELECTIVE]:
			return MessageParser.parse_frc(message, gwid)
		if message_type == Messages.SENSOR_FRC:
			return MessageParser.parse_sensor_frc(message['data']['rsp'], gwid)
		return MessageParser.parse_sensor_read(message['data']['rsp'], gwid)

	@staticmethod
	def parse_charger(message: dict, gwid: str) -> List[Measurement]:
		"""
		Parse Daemon Charger data response message and return list of measurement objects.

		Parameters
		----------
		message: dict
			response message
		gwid: str
			gateway id

		Returns
		-------
		List[Measurements]
			list of measurement objects
		"""
		measurements = []
		device = DeviceInfo(gwid)
		for measurement in ChargerData:
			measurements.append(Measurement(device, f'charger_{measurement.value}', message[measurement.value]))
		return measurements

	@staticmethod
	def parse_frc(message: dict, gwid: str) -> List[Measurement]:
		"""
		Parse Daemon API FRC send response message and return list of measurement objects.

		Parameters
		----------
		message: dict
			response message
		gwid: str
			gateway id

		Returns
		-------
		List[Measurements]
			list of measurement objects
		"""
		measurements = []
		if message['data']['msgId'] != 'FRC_MemR_RSSI':
			return measurements
		data: List[int] = message['data']['rsp']['result']['frcData']
		for address, value in enumerate(data):
			if value == 0:
				continue
			name = 'lastrssi'
			value = float(-abs(value))
			device = DeviceInfo(gwid, address, 0)
			measurements.append(Measurement(device, name, value))
		return measurements

	@staticmethod
	def parse_sensor_frc(message: dict, gwid: str) -> List[Measurement]:
		"""
		Parse Daemon API Sensor FRC response message and return list of measurement objects.

		Parameters
		----------
		message: dict
			response message
		gwid: str
			gateway id

		Returns
		-------
		List[Measurements]
			list of measurement objects
		"""
		measurements = []
		index = 0
		for item in message['result']['sensors']:
			if 'sensor' not in item or item['sensor'] is None:
				continue
			address = int(item['nAdr'])
			sensor = item['sensor']
			try:
				parsed = MessageParser.parse_sensor_object(sensor)
				device = DeviceInfo(gwid, address, index)
				for item in parsed:
					measurements.append(Measurement(device, item['name'], item['value']))
			except (ValueError, SensorException):
				continue
		return measurements

	@staticmethod
	def parse_sensor_read(message: dict, gwid: str) -> List[Measurement]:
		"""
		Parse Daemon API Sensor Read With Types response message and return list of measurement objects.

		Parameters
		----------
		message: dict
			response message
		gwid: str
			gateway id

		Returns
		-------
		List[Measurements]
			list of measurement objects
		"""
		measurements = []
		address = message['nAdr']
		index = 0
		for sensor in message['result']['sensors']:
			try:
				parsed = MessageParser.parse_sensor_object(sensor)
				device = DeviceInfo(gwid, address, index)
				for item in parsed:
					measurements.append(Measurement(device, item['name'], item['value']))
			except (ValueError, SensorException):
				pass
			finally:
				index += 1
		return measurements

	@staticmethod
	def parse_sensor_object(sensor: dict) -> List[Union[Tuple[str, int], Tuple[str, float]]]:
		"""
		Parse Daemon API response sensor object and return sensor name and measured value.

		Parameters
		----------
		sensor: dict
			sensor object

		Returns
		-------
		List[Union[Tuple[str, int], Tuple[str, float]]]
			List of tuples containing sensor name and measured value

		Exceptions
		----------
		SensorException
			raised if value property is missing in sensor object
		"""
		parsed = []
		if 'value' not in sensor or sensor['value'] is None:
			raise SensorException('Value property not in sensor object.')
		if sensor['type'] == 129 or sensor['type'] == 160:
			if 'breakdown' in sensor and sensor['breakdown'] is not None:
				for breakdown in sensor['breakdown']:
					if 'value' in breakdown and breakdown['value'] is not None:
						try:
							parsed.append({
								'name': breakdown['name'].lower(),
								'value': float(breakdown['value'])
							})
						except ValueError:
							continue
			else:
				parsed.append({
					'name': sensor['name'].lower(),
					'value': int(sensor['value'])
				})
		elif sensor['type'] == 192:
			raise SensorException('Data block data not supported.')
		else:
			parsed.append({
				'name': sensor['name'].lower(),
				'value': float(sensor['value'])
			})
		return parsed
