"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import os

from json.decoder import JSONDecodeError
from typing import List, Tuple, Optional
from jsonschema import Draft7Validator
from jsonschema.validators import RefResolver
from paho.mqtt.client import Client as PahoMqttClient, MQTTMessage

from .Enums.Messages import Messages
from .Exceptions.IOException import IOException
from .Exceptions.JSONException import JSONException
from .Exceptions.JSONSchemaException import JSONSchemaException
from .Exceptions.MessageException import MessageException
from .InfluxClientInterface import InfluxClientInterface
from .Logging import Logging
from .MessageParser import MessageParser


class MqttClient:
	"""
	MQTT client class.

	Attributes
	----------
	logger: Logger
		logger object
	influx_client: InfluxClientInterface
		influxdb client
	schema_path: str
		path to message json schema file
	validator: Draft7Validator
		jsonschema draft 7 validator
	host: str
		mqtt broker hostname
	port: int
		mqtt broker port
	user: str
		mqtt broker access username
	password: str
		mqtt broker access password
	client_id: st
		mqtt client id
	topics: List[str]
		list of subscribe topics
	topic_idx: int
		topic index

	Callbacks
	---------
	on_connect(client, userdata, flags, rc)
		Callback function for mqtt.client connect method.
	on_disconnect(self, client, userdata, rc)
		Callback function for mqtt.client disconnect method.
	on_subscribe(self, client, userdata, mid, qos)
		Callback function for mqtt.client subscribe method.
	on_message(self, client, userdata, msg)
		Callback function for processing incoming messages.

	Methods
	-------
	initialize_validator() -> None:
		Initialize jsonschema validator.
	validate_message(message: dict) -> None:
		Validate incoming response message.
	start() -> None:
		Configure MQTT client and start listening loop.
	"""

	def __init__(self, config: dict, influx_client: InfluxClientInterface, schema_path: str) -> None:
		"""
		Construct attributes for MQTT client object.

		Parameters
		----------
		config: dict
			mqtt client configuration dictionary
		influx_client: InfluxClientInterface
			influxdb client object
		schema_path: str
			path to message json schema file

		Returns
		-------
		None
		"""
		self.logger = Logging.get_logger()
		self.influx_client: InfluxClientInterface = influx_client
		self.schema_path: str = schema_path
		self.validator: Optional[Draft7Validator] = None
		self.host: str = config['host']
		self.port: int = config['port']
		self.user: str = config['user']
		self.password: str = config['password']
		self.client_id: str = f'{config["client"]}_{str(os.getpid())}'
		self.topics: List[str] = config['topics']
		self.topic_idx: int = 0
		self.client = PahoMqttClient(self.client_id)

	def initialize_validator(self) -> None:
		"""
		Initialize jsonschema validator.

		Parameters
		----------
		None

		Returns
		-------
		None

		Exceptions
		----------
		IOException
			raised if schema file does not exist
		JSONDecodeError
			raised if schema file does not have valid json syntax
		"""
		self.logger.info('Loaded request validating schema from %s.', self.schema_path)
		try:
			with open(self.schema_path, mode='r', encoding='utf-8') as file:
				schema = json.load(file)
				resolver = RefResolver.from_schema(schema)
				self.validator = Draft7Validator(schema, resolver=resolver)
		except IOError as io_error:
			error_str = Logging.io_error(io_error)
			self.logger.error(error_str)
			raise IOException(error_str) from io_error
		except JSONDecodeError as json_error:
			error_str = Logging.json_file_error(json_error, self.schema_path)
			self.logger.error(error_str)
			raise JSONException(error_str) from json_error

	def validate_message(self, message: dict) -> None:
		"""
		Validate incoming response message.

		Parameters
		----------
		message: dict
			response message from mqtt broker

		Returns
		-------
		None

		Exceptions
		----------
		JSONSchemaException
			raised if received message does not validate against message schema
		"""
		schema_errors = sorted(self.validator.iter_errors(message), key=str)
		if schema_errors:
			errors = []
			for error in schema_errors:
				errors.append(error.message)
			raise JSONSchemaException('\n'.join(errors))

	def on_connect(self, client, userdata: dict, flags, connection_state: int) -> None:
		"""
		Process on_connect events.

		Subscribes to specified MQTT broker topics.

		Parameters
		----------
		client
			mqtt.client object
		userdata: dict
			private client user data
		flags
			response flags from MQTT broker
		connection_state: int
			return code

		Returns
		-------
		None
		"""
		# pylint: disable=unused-argument
		if connection_state != 0:
			self.logger.info('Failed to connect to %s:%d with code %d', self.host, self.port, connection_state)
			return
		self.logger.info('Connected to %s:%d.', self.host, self.port)
		for topic in self.topics:
			self.client.subscribe(topic)

	def on_disconnect(self, client, userdata: dict, connection_state: int) -> None:
		"""
		Process disconnect and connection lost events.

		Logs disconnect and connection lost event.

		Parameters
		----------
		client
			mqtt.client object
		userdata: dict
			private client user data
		connection_state: int
			return code

		Returns
		-------
		None
		"""
		# pylint: disable=unused-argument
		self.logger.info('Disconnected with code %d, reconnecting...', connection_state)
		self.topic_idx = 0

	def on_subscribe(self, client, userdata: dict, mid: int, qos: Tuple[int]) -> None:
		"""
		Process on_subscribe events.

		Parameters
		----------
		client
			mqtt.client object
		userdata: dict
			private client user data
		mid: int
			message id
		qos: Tuple[int]
			quality of service

		Returns
		-------
		None
		"""
		# pylint: disable=unused-argument
		self.logger.info('Subscribed to topic %s with qos %d.', self.topics[self.topic_idx], qos[0])
		self.topic_idx += 1

	def on_message(self, client, userdata: dict, msg: MQTTMessage) -> None:
		"""
		Process incoming messages.

		Validates incoming message, if the message is valid, it is passed for processing.

		Parameters
		----------
		client
			mqtt.client object
		userdata: dict
			private client user data
		msg: MQTTMessage
			received message

		Returns
		-------
		None
		"""
		# pylint: disable=unused-argument
		try:
			payload = json.loads(msg.payload.decode('utf-8'))
			message_type = payload['mType']
			if not Messages.exists(message_type):
				raise MessageException(f'Request of type {message_type} is ignored.')
			self.validate_message(payload)
			self.logger.debug('Received message: %s', json.dumps(payload))
			gwid = msg.topic.split("/", 4)[1]
			measurements = MessageParser.parse_message(payload, gwid)
			self.influx_client.write(measurements)
		except JSONDecodeError as decode_error:
			self.logger.warning(Logging.json_response_parse_error(decode_error))
		except JSONSchemaException as validation_error:
			self.logger.debug(validation_error.message)
		except MessageException as message_error:
			self.logger.debug(message_error.message)

	def start(self) -> None:
		"""
		Configure MQTT client and start listening loop.

		Parameters
		----------
		None

		Returns
		-------
		None
		"""
		self.initialize_validator()
		self.client.username_pw_set(self.user, self.password)
		self.client.on_connect = self.on_connect
		self.client.on_disconnect = self.on_disconnect
		self.client.on_subscribe = self.on_subscribe
		self.client.on_message = self.on_message
		self.client.connect(self.host, self.port)
		self.client.loop_forever()
