"""
Copyright 2021-2023 MICRORISC s.r.o.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=C0103

import argparse
import sys
from logging import Logger
import sentry_sdk

from sentry_sdk.integrations.atexit import AtexitIntegration

from . import __version__
from .Config import Config
try:
	from .InfluxClient import InfluxClient
	influxdb_new = True
except ModuleNotFoundError:
	influxdb_new = False
try:
	from .InfluxClientLegacy import InfluxClientLegacy
	influxdb_legacy = True
except ModuleNotFoundError:
	influxdb_legacy = False
from .Logging import Logging
from .MqttClient import MqttClient
from .Enums.ErrorCodes import ErrorCodes
from .Exceptions.BridgeBaseException import BridgeBaseException


def main():
	"""Influxdb Bridge entry point."""
	parser = argparse.ArgumentParser(prog='IQRF Gateway Influxdb Bridge', description='MQTT to Influxdb bridge utility.')
	parser.add_argument('-c', '--config', type=str, dest='config',
						default='/etc/iqrf-gateway-influxdb-bridge/config.json',
						help='Path to configuration file.', required=False)
	parser.add_argument('-s', '--schema-config', type=str, dest='config_schema',
						default='/usr/share/iqrf-gateway-influxdb-bridge/schema_config.json',
						help='Path to configuration schema file.', required=False)
	parser.add_argument('-m', '--schema-message', type=str, dest='message_schema',
						default='/usr/share/iqrf-gateway-influxdb-bridge/schema_message.json',
						help='Path to message schema file.', required=False)
	parser.add_argument('-l', '--legacy', dest='legacy', action='store_true',
						help='Use legacy client for communication with Influxdb 1.x.')
	parser.add_argument('-v', '--version', action='version', version='%(prog)s v' + __version__)
	args = parser.parse_args()

	sentry_sdk.init(
		dsn='https://9ed398f2e0c9428cacf1efc4027b2fa3@sentry.iqrf.org/13',
		traces_sample_rate=1.0,
		integrations=[
			AtexitIntegration(),
		],
	)

	logger: Logger = Logging.get_logger()
	logger.info('IQRF Gateway Influxdb Bridge v%s', __version__)

	try:
		config = Config.read(args.config, args.config_schema)
		log_level = config['logLevel']
		Logging.set_level(log_level)
		logger.setLevel(Logging.level_name_to_value(log_level))
		if args.legacy:
			if not influxdb_legacy:
				return ErrorCodes.MISSING_INFLUXDB_CLIENT_ERROR
			influx_client = InfluxClientLegacy(config['influx'])
		else:
			if not influxdb_new:
				return ErrorCodes.MISSING_INFLUXDB_CLIENT_ERROR
			influx_client = InfluxClient(config['influx'])
		influx_client.start()
		mqtt_client = MqttClient(config['mqtt'], influx_client, args.message_schema)
		mqtt_client.start()
	except BridgeBaseException as error:
		return error.code

	return ErrorCodes.OK


if __name__ == '__main__':
	sys.exit(main())
